"""
https://github.github.com/gfm/#atx-headings
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_atx_headings_extra_1():
    """
    Test case extra 1:  ATX headings starts with a backslash escape
    """

    # Arrange
    source_markdown = """## \\\\this is a fun day"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):\\\b\\this is a fun day: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>\\this is a fun day</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_2():
    """
    Test case extra 2:  ATX headings starts with a backslash as in a hard line break
    """

    # Arrange
    source_markdown = """## \\"""
    expected_tokens = ["[atx(1,1):2:0:]", "[text(1,4):\\: ]", "[end-atx::]"]
    expected_gfm = """<h2>\\</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_3():
    """
    Test case extra 3:  ATX headings starts with 2+ spaces as in a hard line break
    """

    # Arrange
    source_markdown = """##    """
    expected_tokens = ["[atx(1,1):2:0:]", "[text(1,7)::    ]", "[end-atx::]"]
    expected_gfm = """<h2></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_4():
    """
    Test case extra 4:  ATX headings string starting with a code span.
    """

    # Arrange
    source_markdown = """## ``this`` is a fun day"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[icode-span(1,4):this:``::]",
        "[text(1,12): is a fun day:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><code>this</code> is a fun day</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_5():
    """
    Test case extra 5:  ATX headings string starting with a character reference.
    """

    # Arrange
    source_markdown = """## &amp; the band played on"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):\a&amp;\a\a&\a&amp;\a\a the band played on: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>&amp; the band played on</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_6():
    """
    Test case extra 6:  ATX headings string starting with a raw html block.
    """

    # Arrange
    source_markdown = """## <there it='is'>, really"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[raw-html(1,4):there it='is']",
        "[text(1,19):, really:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><there it='is'>, really</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_7():
    """
    Test case extra 7:  ATX headings string starting with an URI autolink
    """

    # Arrange
    source_markdown = """## <http://www.google.com> is where to look"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[uri-autolink(1,4):http://www.google.com]",
        "[text(1,27): is where to look:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><a href="http://www.google.com">http://www.google.com</a> is where to look</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_8():
    """
    Test case extra 8:  ATX headings string starting with an email autolink
    """

    # Arrange
    source_markdown = """## <foo@bar.com> for more information"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[email-autolink(1,4):foo@bar.com]",
        "[text(1,17): for more information:]",
        "[end-atx::]",
    ]
    expected_gfm = (
        """<h2><a href="mailto:foo@bar.com">foo@bar.com</a> for more information</h2>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_9():
    """
    Test case extra 9:  ATX headings string starting with an emphasis
    """

    # Arrange
    source_markdown = """## *it's* me!"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[emphasis(1,4):1:*]",
        "[text(1,5):it's:]",
        "[end-emphasis(1,9)::]",
        "[text(1,10): me!:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><em>it's</em> me!</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_10():
    """
    Test case extra 10:  ATX headings string starting with a link.  also see 183
    """

    # Arrange
    source_markdown = """## [Foo](/uri) is a link"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[link(1,4):inline:/uri:::::Foo:False::::]",
        "[text(1,5):Foo:]",
        "[end-link::]",
        "[text(1,15): is a link:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><a href="/uri">Foo</a> is a link</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_11():
    """
    Test case extra 11:  ATX headings string starting with an image
    """

    # Arrange
    source_markdown = """## ![foo](/url "title") is an image"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        '[image(1,4):inline:/url:title:foo::::foo:False:":: :]',
        "[text(1,24): is an image:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><img src="/url" alt="foo" title="title" /> is an image</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_12():
    """
    Test case extra 12:  ATX headings containing a backslash
    """

    # Arrange
    source_markdown = """## this is a \\\\fun\\\\ day"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):this is a \\\b\\fun\\\b\\ day: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>this is a \\fun\\ day</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_13():
    """
    Test case extra 13:  ATX headings containing a code span.
    """

    # Arrange
    source_markdown = """## this is a ``fun`` day"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):this is a : ]",
        "[icode-span(1,14):fun:``::]",
        "[text(1,21): day:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>this is a <code>fun</code> day</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_14():
    """
    Test case extra 14:  ATX headings containing a character reference.
    """

    # Arrange
    source_markdown = """## fun &amp; joy"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):fun \a&amp;\a\a&\a&amp;\a\a joy: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>fun &amp; joy</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_15():
    """
    Test case extra 15:  ATX heading containing a raw html block.
    """

    # Arrange
    source_markdown = """## where <there it='is'> it"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):where : ]",
        "[raw-html(1,10):there it='is']",
        "[text(1,25): it:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>where <there it='is'> it</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_16():
    """
    Test case extra 16:  ATX heading containing an URI autolink
    """

    # Arrange
    source_markdown = """## look at <http://www.google.com> for answers"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):look at : ]",
        "[uri-autolink(1,12):http://www.google.com]",
        "[text(1,35): for answers:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>look at <a href="http://www.google.com">http://www.google.com</a> for answers</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_17():
    """
    Test case extra 17:  ATX heading containing an email autolink
    """

    # Arrange
    source_markdown = """## email <foo@bar.com> for answers"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):email : ]",
        "[email-autolink(1,10):foo@bar.com]",
        "[text(1,23): for answers:]",
        "[end-atx::]",
    ]
    expected_gfm = (
        """<h2>email <a href="mailto:foo@bar.com">foo@bar.com</a> for answers</h2>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_18():
    """
    Test case extra 18:  ATX heading containing emphasis
    """

    # Arrange
    source_markdown = """## really! *it's me!* here!"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):really! : ]",
        "[emphasis(1,12):1:*]",
        "[text(1,13):it's me!:]",
        "[end-emphasis(1,21)::]",
        "[text(1,22): here!:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>really! <em>it's me!</em> here!</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_19():
    """
    Test case extra 19:  ATX heading containing a link.
    """

    # Arrange
    source_markdown = """## look at [Foo](/uri) for more"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):look at : ]",
        "[link(1,12):inline:/uri:::::Foo:False::::]",
        "[text(1,13):Foo:]",
        "[end-link::]",
        "[text(1,23): for more:]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>look at <a href="/uri">Foo</a> for more</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_20():
    """
    Test case extra 20:  ATX heading containing an image
    """

    # Arrange
    source_markdown = """## special ![foo](/url "title") headings"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):special : ]",
        '[image(1,12):inline:/url:title:foo::::foo:False:":: :]',
        "[text(1,32): headings:]",
        "[end-atx::]",
    ]
    expected_gfm = (
        """<h2>special <img src="/url" alt="foo" title="title" /> headings</h2>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_21():
    """
    Test case extra 21:  ATX headings ends with a backslash escape
    """

    # Arrange
    source_markdown = """## this is a fun day\\\\"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):this is a fun day\\\b\\: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>this is a fun day\\</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_22():
    """
    Test case extra 22:  ATX headings ends with a backslash as in a hard line break
    """

    # Arrange
    source_markdown = """## this was \\"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):this was \\: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>this was \\</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_23():
    """
    Test case extra 23:  ATX headings ends with 2+ spaces as in a hard line break
    """

    # Arrange
    source_markdown = """## what? no line break?   """
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):what? no line break?: ]",
        "[end-atx:   :]",
    ]
    expected_gfm = """<h2>what? no line break?</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_24():
    """
    Test case extra 24:  ATX headings string ending with a code span.
    """

    # Arrange
    source_markdown = """## this is a fun ``day``"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):this is a fun : ]",
        "[icode-span(1,18):day:``::]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>this is a fun <code>day</code></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_25():
    """
    Test case extra 25:  ATX headings string ending with a character reference.
    """

    # Arrange
    source_markdown = """## the band played on &amp;"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):the band played on \a&amp;\a\a&\a&amp;\a\a: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>the band played on &amp;</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_26():
    """
    Test case extra 26:  ATX headings string ending with a raw html block.
    """

    # Arrange
    source_markdown = """## really, <there it='is'>"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):really, : ]",
        "[raw-html(1,12):there it='is']",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>really, <there it='is'></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_27():
    """
    Test case extra 27:  ATX headings string ending with an URI autolink
    """

    # Arrange
    source_markdown = """## look at <http://www.google.com>"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):look at : ]",
        "[uri-autolink(1,12):http://www.google.com]",
        "[end-atx::]",
    ]
    expected_gfm = (
        """<h2>look at <a href="http://www.google.com">http://www.google.com</a></h2>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_28():
    """
    Test case extra 28:  ATX headings string ending with an email autolink
    """

    # Arrange
    source_markdown = """## for more information, contact <foo@bar.com>"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):for more information, contact : ]",
        "[email-autolink(1,34):foo@bar.com]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>for more information, contact <a href="mailto:foo@bar.com">foo@bar.com</a></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_29():
    """
    Test case extra 29:  ATX headings string ending with an emphasis
    """

    # Arrange
    source_markdown = """## it's *me*"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):it's : ]",
        "[emphasis(1,9):1:*]",
        "[text(1,10):me:]",
        "[end-emphasis(1,12)::]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>it's <em>me</em></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_30():
    """
    Test case extra 30:  ATX headings string ending with a link.
    """

    # Arrange
    source_markdown = """## a link looks like [Foo](/uri)"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):a link looks like : ]",
        "[link(1,22):inline:/uri:::::Foo:False::::]",
        "[text(1,23):Foo:]",
        "[end-link::]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>a link looks like <a href="/uri">Foo</a></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_31():
    """
    Test case extra 31:  ATX headings string ending with an image
    """

    # Arrange
    source_markdown = """## an image is ![foo](/url "title")"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):an image is : ]",
        '[image(1,16):inline:/url:title:foo::::foo:False:":: :]',
        "[end-atx::]",
    ]
    expected_gfm = """<h2>an image is <img src="/url" alt="foo" title="title" /></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_32():
    """
    Test case extra 32:  ATX heading this is only a backslash escape
    """

    # Arrange
    source_markdown = """## \\\\"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):\\\b\\: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>\\</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_33():
    """
    Test case extra 33:  ATX heading this is only a backslash as in a hard line break
    """

    # Arrange
    source_markdown = """## \\"""
    expected_tokens = ["[atx(1,1):2:0:]", "[text(1,4):\\: ]", "[end-atx::]"]
    expected_gfm = """<h2>\\</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_34():
    """
    Test case extra 34:  ATX heading this is only 2+ spaces as in a hard line break
    """

    # Arrange
    source_markdown = """##    """
    expected_tokens = ["[atx(1,1):2:0:]", "[text(1,7)::    ]", "[end-atx::]"]
    expected_gfm = """<h2></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_35():
    """
    Test case extra 35:  ATX heading this is only a code span.
    """

    # Arrange
    source_markdown = """## ``day``"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[icode-span(1,4):day:``::]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><code>day</code></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_36():
    """
    Test case extra 36:  ATX heading this is only a character reference.
    """

    # Arrange
    source_markdown = """## &amp;"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4):\a&amp;\a\a&\a&amp;\a\a: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2>&amp;</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_37():
    """
    Test case extra 37:  ATX heading this is only a raw html block.
    """

    # Arrange
    source_markdown = """## <there it='is'>"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[raw-html(1,4):there it='is']",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><there it='is'></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_38():
    """
    Test case extra 38:  ATX heading this is only an URI autolink
    """

    # Arrange
    source_markdown = """## <http://www.google.com>"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[uri-autolink(1,4):http://www.google.com]",
        "[end-atx::]",
    ]
    expected_gfm = (
        """<h2><a href="http://www.google.com">http://www.google.com</a></h2>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_39():
    """
    Test case extra 39:  ATX heading this is only an email autolink
    """

    # Arrange
    source_markdown = """## <foo@bar.com>"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[email-autolink(1,4):foo@bar.com]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><a href="mailto:foo@bar.com">foo@bar.com</a></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_40():
    """
    Test case extra 40:  ATX heading this is only an emphasis
    """

    # Arrange
    source_markdown = """## *me*"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[emphasis(1,4):1:*]",
        "[text(1,5):me:]",
        "[end-emphasis(1,7)::]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><em>me</em></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_41():
    """
    Test case extra 41:  ATX heading this is only a link.
    """

    # Arrange
    source_markdown = """## [Foo](/uri)"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        "[link(1,4):inline:/uri:::::Foo:False::::]",
        "[text(1,5):Foo:]",
        "[end-link::]",
        "[end-atx::]",
    ]
    expected_gfm = """<h2><a href="/uri">Foo</a></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_42():
    """
    Test case extra 42:  ATX heading this is only an image
    """

    # Arrange
    source_markdown = """## ![foo](/url "title")"""
    expected_tokens = [
        "[atx(1,1):2:0:]",
        "[text(1,4)::\a \a\x03\a]",
        '[image(1,4):inline:/url:title:foo::::foo:False:":: :]',
        "[end-atx::]",
    ]
    expected_gfm = """<h2><img src="/url" alt="foo" title="title" /></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_43x():
    """
    Test case extra 43x:  Not quote ATX.
    """

    # Arrange
    source_markdown = """#Heading 1 with no blank lines"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):#Heading 1 with no blank lines:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>#Heading 1 with no blank lines</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_43a():
    """
    Test case extra 43:  Not quote ATX.
    """

    # Arrange
    source_markdown = """#Heading 1 with no blank lines
##Heading 2 with no blank lines"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):#Heading 1 with no blank lines\n##Heading 2 with no blank lines::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>#Heading 1 with no blank lines\n##Heading 2 with no blank lines</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_44():
    """
    Test case extra 44:  Tab character instead of space between the # and text.
    """

    # Arrange
    source_markdown = """#\tHeading 1 with no blank lines"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):#    Heading 1 with no blank lines:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>#    Heading 1 with no blank lines</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_atx_headings_extra_44a():
    """
    Test case extra 44:  Tab character instead of space between the # and text.
    """

    # Arrange
    source_markdown = """# Heading 1 with no blank lines\t#"""
    expected_tokens = [
        "[atx(1,1):1:0:]",
        "[text(1,3):Heading 1 with no blank lines\t#: ]",
        "[end-atx::]",
    ]
    expected_gfm = """<h1>Heading 1 with no blank lines\t#</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide processing for the block quotes.
"""
import logging

from pymarkdown.container_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.leaf_block_processor import LeafBlockProcessor
from pymarkdown.parser_helper import ParserHelper, PositionMarker
from pymarkdown.parser_logger import ParserLogger
from pymarkdown.stack_token import (
    BlockQuoteStackToken,
    FencedCodeBlockStackToken,
    HtmlBlockStackToken,
    IndentedCodeBlockStackToken,
    LinkDefinitionStackToken,
    ParagraphStackToken,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class BlockQuoteProcessor:
    """
    Class to provide processing for the block quotes.
    """

    __block_quote_character = ">"

    @staticmethod
    def is_block_quote_start(
        line_to_parse, start_index, extracted_whitespace, adj_ws=None
    ):
        """
        Determine if we have the start of a block quote section.
        """

        if adj_ws is None:
            adj_ws = extracted_whitespace

        return ParserHelper.is_length_less_than_or_equal_to(
            adj_ws, 3
        ) and ParserHelper.is_character_at_index(
            line_to_parse, start_index, BlockQuoteProcessor.__block_quote_character
        )

    @staticmethod
    def check_for_lazy_handling(
        parser_state,
        this_bq_count,
        stack_bq_count,
        line_to_parse,
        extracted_whitespace,
    ):
        """
        Check if there is any processing to be handled during the handling of
        lazy continuation lines in block quotes.
        """
        POGGER.debug("__check_for_lazy_handling")
        container_level_tokens = []
        POGGER.debug(
            "this_bq_count>$>>stack_bq_count>>$<<",
            this_bq_count,
            stack_bq_count,
        )
        if this_bq_count == 0 and stack_bq_count > 0:
            POGGER.debug("haven't processed")

            is_leaf_block_start = (
                LeafBlockProcessor.is_paragraph_ending_leaf_block_start(
                    parser_state,
                    line_to_parse,
                    0,
                    extracted_whitespace,
                    exclude_thematic_break=True,
                )
            )

            if (
                parser_state.token_stack[-1].is_code_block
                or parser_state.token_stack[-1].is_html_block
                or is_leaf_block_start
            ):
                POGGER.debug("__check_for_lazy_handling>>code block")
                assert not container_level_tokens
                container_level_tokens, _ = parser_state.close_open_blocks_fn(
                    parser_state,
                    only_these_blocks=[
                        BlockQuoteStackToken,
                        type(parser_state.token_stack[-1]),
                    ],
                    include_block_quotes=True,
                    was_forced=True,
                )

        return container_level_tokens

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    @staticmethod
    def handle_block_quote_block(
        parser_state,
        position_marker,
        extracted_whitespace,
        adj_ws,
        this_bq_count,
        stack_bq_count,
    ):
        """
        Handle the processing of a block quote block.
        """
        POGGER.debug("handle_block_quote_block>>start")

        (
            did_process,
            was_container_start,
            avoid_block_starts,
            did_blank,
            removed_chars_at_start,
            last_block_quote_index,
            end_of_bquote_start_index,
            text_removed_by_container,
            requeue_line_info,
            leaf_tokens,
            container_level_tokens,
        ) = (False, False, False, False, 0, 0, -1, None, None, [], [])

        adjusted_text_to_parse, adjusted_index_number = (
            position_marker.text_to_parse,
            position_marker.index_number,
        )

        POGGER.debug(
            "handle_block_quote_block>>text>:$:<", position_marker.text_to_parse
        )
        POGGER.debug(
            "handle_block_quote_block>>extracted_whitespace>:$:<",
            extracted_whitespace,
        )
        POGGER.debug("handle_block_quote_block>>adj_ws>:$:<", adj_ws)
        POGGER.debug(
            "handle_block_quote_block>>was_link_definition_started>:$:<",
            parser_state.token_stack[-1].was_link_definition_started,
        )
        if BlockQuoteProcessor.is_block_quote_start(
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
            adj_ws=adj_ws,
        ):
            POGGER.debug("handle_block_quote_block>>block-start")
            (
                adjusted_text_to_parse,
                adjusted_index_number,
                leaf_tokens,
                container_level_tokens,
                stack_bq_count,
                alt_this_bq_count,
                removed_chars_at_start,
                did_blank,
                last_block_quote_index,
                text_removed_by_container,
                avoid_block_starts,
                requeue_line_info,
            ) = BlockQuoteProcessor.__handle_block_quote_section(
                parser_state,
                position_marker,
                stack_bq_count,
                extracted_whitespace,
            )
            POGGER.debug(">>avoid_block_starts>>$", avoid_block_starts)

            # TODO for nesting, may need to augment with this_bq_count already set.
            this_bq_count = alt_this_bq_count
            if not this_bq_count == 0:
                POGGER.debug(">>>>>>>>>>>>>>>$>>>$", this_bq_count, alt_this_bq_count)

            if last_block_quote_index != -1:
                did_process, end_of_bquote_start_index = True, adjusted_index_number
            else:
                did_process, end_of_bquote_start_index = False, -1
            was_container_start = did_process
        elif parser_state.token_stack[-1].was_link_definition_started:
            stack_index = parser_state.find_last_block_quote_on_stack()
            if stack_index > 0:
                last_block_token = parser_state.token_stack[
                    stack_index
                ].matching_markdown_token
                POGGER.debug(
                    "handle_block w/ no open>>found>>$",
                    last_block_token,
                )
                last_block_token.add_leading_spaces("")

        POGGER.debug("handle_block_quote_block>>end>>did_process>>$", did_process)
        return (
            did_process,
            was_container_start,
            end_of_bquote_start_index,
            this_bq_count,
            stack_bq_count,
            adjusted_text_to_parse,
            adjusted_index_number,
            leaf_tokens,
            container_level_tokens,
            removed_chars_at_start,
            did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            requeue_line_info,
        )

    # pylint: enable=too-many-arguments
    # pylint: enable=too-many-locals

    @staticmethod
    def __count_block_quote_starts(
        line_to_parse,
        start_index,
        stack_bq_count,
        is_top_of_stack_fenced_code_block,
        is_top_of_stack_is_html_block,
    ):
        """
        Having detected a block quote character (">") on a line, continue to consume
        and count while the block quote pattern is there.
        """

        (
            osi,
            oltp,
            this_bq_count,
            last_block_quote_index,
            avoid_block_starts,
            adjusted_line,
        ) = (
            start_index,
            line_to_parse[:],
            0,
            -1,
            False,
            line_to_parse,
        )
        if stack_bq_count == 0 and is_top_of_stack_fenced_code_block:
            start_index -= 1
        else:
            this_bq_count += 1
            start_index += 1
            last_block_quote_index = start_index

            POGGER.debug(
                "stack_bq_count--$--is_top_of_stack_fenced_code_block--$",
                stack_bq_count,
                is_top_of_stack_fenced_code_block,
            )

            while True:
                if ParserHelper.is_character_at_index_whitespace(
                    adjusted_line, start_index
                ):
                    if adjusted_line[start_index] == ParserHelper.tab_character:
                        adjusted_tab_length = ParserHelper.calculate_length(
                            ParserHelper.tab_character, start_index=start_index
                        )
                        POGGER.debug("adj--$--", adjusted_line)
                        parts = [
                            adjusted_line[0:start_index],
                            ParserHelper.repeat_string(
                                ParserHelper.space_character, adjusted_tab_length
                            ),
                            adjusted_line[start_index + 1 :],
                        ]
                        adjusted_line = "".join(parts)
                        POGGER.debug("--$--", adjusted_line)
                    start_index += 1

                POGGER.debug(
                    "this_bq_count--$--stack_bq_count--$--is_top_of_stack_is_html_block--$",
                    this_bq_count,
                    stack_bq_count,
                    is_top_of_stack_is_html_block,
                )
                if is_top_of_stack_is_html_block:
                    if this_bq_count == stack_bq_count:
                        POGGER.debug(
                            "block quote levels don't increase during html block, ignoring"
                        )
                        avoid_block_starts = ParserHelper.is_character_at_index(
                            adjusted_line,
                            start_index,
                            BlockQuoteProcessor.__block_quote_character,
                        )
                        POGGER.debug("avoid_block_starts=$", avoid_block_starts)
                        break
                    if this_bq_count > stack_bq_count:
                        start_index, adjusted_line, last_block_quote_index = (
                            osi,
                            oltp,
                            -1,
                        )
                        avoid_block_starts, this_bq_count = True, stack_bq_count
                        break

                if is_top_of_stack_fenced_code_block and (
                    this_bq_count >= stack_bq_count
                ):
                    break

                if start_index == len(
                    adjusted_line
                ) or ParserHelper.is_character_at_index_not(
                    adjusted_line,
                    start_index,
                    BlockQuoteProcessor.__block_quote_character,
                ):
                    break
                this_bq_count += 1
                start_index += 1
                last_block_quote_index = start_index

            POGGER.debug(
                "__count_block_quote_starts--$--$--",
                start_index,
                adjusted_line,
            )
        return (
            this_bq_count,
            start_index,
            adjusted_line,
            last_block_quote_index,
            avoid_block_starts,
        )

    # pylint: disable=too-many-locals, too-many-statements
    @staticmethod
    def __handle_block_quote_section(
        parser_state,
        position_marker,
        stack_bq_count,
        extracted_whitespace,
    ):
        """
        Handle the processing of a section clearly identified as having block quotes.
        """
        line_to_parse, start_index = (
            position_marker.text_to_parse,
            position_marker.index_number,
        )

        text_removed_by_container, did_blank, removed_chars_at_start = None, False, 0
        leaf_tokens, container_level_tokens, original_start_index = [], [], start_index

        POGGER.debug(
            "IN>__handle_block_quote_section---$<<<",
            line_to_parse,
        )
        POGGER.debug("stack_bq_count--$", stack_bq_count)
        POGGER.debug("token_stack[-1]--$", parser_state.token_stack[-1])

        POGGER.debug(
            "__handle_block_quote_section---$--$--",
            start_index,
            line_to_parse,
        )
        (
            this_bq_count,
            start_index,
            line_to_parse,
            last_block_quote_index,
            avoid_block_starts,
        ) = BlockQuoteProcessor.__count_block_quote_starts(
            line_to_parse,
            start_index,
            stack_bq_count,
            parser_state.token_stack[-1].is_fenced_code_block,
            parser_state.token_stack[-1].is_html_block,
        )
        POGGER.debug(">>start_index>>$", start_index)
        POGGER.debug(">>original_start_index>>$", original_start_index)
        if last_block_quote_index == -1:
            POGGER.debug("BAIL")
            return (
                line_to_parse,
                start_index,
                leaf_tokens,
                container_level_tokens,
                stack_bq_count,
                this_bq_count,
                removed_chars_at_start,
                did_blank,
                last_block_quote_index,
                text_removed_by_container,
                avoid_block_starts,
                None,
            )

        POGGER.debug(">>avoid_block_starts>>$", avoid_block_starts)
        POGGER.debug(
            "__handle_block_quote_section---this_bq_count--$--$--$--",
            this_bq_count,
            start_index,
            line_to_parse,
        )
        POGGER.debug(
            "ORIG-->WS[$]--SI[$]--[$]",
            extracted_whitespace,
            original_start_index,
            parser_state.original_line_to_parse,
        )
        POGGER.debug("NOW -->SI[$]--[$]", start_index, line_to_parse)

        if not parser_state.token_stack[-1].is_fenced_code_block:
            POGGER.debug("handle_block_quote_section>>not fenced")
            (
                container_level_tokens,
                stack_bq_count,
                requeue_line_info,
            ) = BlockQuoteProcessor.__ensure_stack_at_level(
                parser_state,
                this_bq_count,
                stack_bq_count,
                extracted_whitespace,
                position_marker,
                original_start_index,
            )
            if requeue_line_info:
                return (
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    requeue_line_info,
                )

            removed_text = f"{extracted_whitespace}{line_to_parse[position_marker.index_number : start_index]}"
            POGGER.debug(
                "==EWS[$],OSI[$],SI[$],LTP[$]",
                extracted_whitespace,
                original_start_index,
                position_marker.index_number,
                position_marker.text_to_parse,
            )
            line_to_parse, removed_chars_at_start = (
                line_to_parse[start_index:],
                start_index,
            )
            POGGER.debug("==REM[$],LTP[$]", removed_text, line_to_parse)

            stack_index, text_removed_by_container = (
                parser_state.find_last_block_quote_on_stack(),
                removed_text,
            )
            assert stack_index != -1
            found_bq_stack_token = parser_state.token_stack[stack_index]
            found_bq_stack_token.matching_markdown_token.add_leading_spaces(
                removed_text
            )

            if not line_to_parse.strip():
                POGGER.debug("call __handle_block_quote_section>>handle_blank_line")

                POGGER.debug("__hbqs>>this_bq_count>>$", this_bq_count)
                POGGER.debug("__hbqs>>stack_bq_count>>$", stack_bq_count)

                POGGER.debug("__hbqs>>token_stack>>$", parser_state.token_stack)
                possible_list_start_index, forced_close_until_index = (
                    this_bq_count + 1,
                    None,
                )
                if (
                    possible_list_start_index < len(parser_state.token_stack)
                    and parser_state.token_stack[possible_list_start_index].is_list
                ):
                    forced_close_until_index = possible_list_start_index
                    POGGER.debug(
                        "__hbqs>>fgg>>$<<",
                        parser_state.token_stack[possible_list_start_index],
                    )
                POGGER.debug("__hbqs>>fgg>>$", forced_close_until_index)

                adjusted_position_marker = PositionMarker(
                    position_marker.line_number,
                    len(text_removed_by_container),
                    position_marker.text_to_parse,
                )
                did_blank = True
                (leaf_tokens, requeue_line_info) = parser_state.handle_blank_line_fn(
                    parser_state,
                    line_to_parse,
                    from_main_transform=False,
                    forced_close_until_index=forced_close_until_index,
                    position_marker=adjusted_position_marker,
                )
                assert not (requeue_line_info and requeue_line_info.lines_to_requeue)
        else:
            POGGER.debug("handle_block_quote_section>>fenced")
            assert start_index >= 0
            removed_text, line_to_parse = (
                line_to_parse[0:start_index],
                line_to_parse[start_index:],
            )

            POGGER.debug("__hbqs>>removed_text>>$", removed_text)
            POGGER.debug("__hbqs>>line_to_parse>>$", line_to_parse)
            POGGER.debug("__hbqs>>this_bq_count>>$", this_bq_count)
            POGGER.debug("__hbqs>>stack_bq_count>>$", stack_bq_count)

            if this_bq_count < stack_bq_count:
                (container_level_tokens, _,) = parser_state.close_open_blocks_fn(
                    parser_state,
                    only_these_blocks=[
                        FencedCodeBlockStackToken,
                    ],
                    was_forced=True,
                )
                stack_bq_count = BlockQuoteProcessor.__decrease_stack_to_level(
                    parser_state, this_bq_count, stack_bq_count, container_level_tokens
                )

            stack_index = parser_state.find_last_block_quote_on_stack()
            found_bq_stack_token = parser_state.token_stack[stack_index]
            POGGER.debug(
                "found_bq_stack_token---$<<<",
                found_bq_stack_token,
            )
            found_bq_stack_token.matching_markdown_token.add_leading_spaces(
                removed_text
            )

        POGGER.debug(
            "OUT>__handle_block_quote_section---$<<<",
            line_to_parse,
        )

        return (
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            stack_bq_count,
            this_bq_count,
            removed_chars_at_start,
            did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            None,
        )

    # pylint: enable=too-many-locals, too-many-statements

    # pylint: disable=too-many-arguments, too-many-statements, too-many-locals
    @staticmethod
    def __ensure_stack_at_level(
        parser_state,
        this_bq_count,
        stack_bq_count,
        extracted_whitespace,
        position_marker,
        original_start_index,
    ):
        """
        Ensure that the block quote stack is at the proper level on the stack.
        """

        container_level_tokens, stack_increase_needed, stack_decrease_needed = (
            [],
            False,
            False,
        )
        POGGER.debug(
            "__ensure_stack_at_level>>this_bq_count>>$>>stack_bq_count>>$",
            this_bq_count,
            stack_bq_count,
        )
        if this_bq_count > stack_bq_count:
            POGGER.debug(
                "__ensure_stack_at_level>>increase to new level",
            )
            stack_increase_needed = True
        elif this_bq_count < stack_bq_count:
            POGGER.debug(
                "__ensure_stack_at_level>>possible decrease to new level",
            )
            top_token_on_stack = parser_state.token_stack[-1]
            POGGER.debug("__ensure_stack_at_level>>$", top_token_on_stack)
            stack_decrease_needed = (
                top_token_on_stack.is_indented_code_block
                or top_token_on_stack.is_html_block
            )
            POGGER.debug(
                "__ensure_stack_at_level>>decrease to new level=$",
                stack_decrease_needed,
            )

        if stack_increase_needed or stack_decrease_needed:
            POGGER.debug(
                "token_stack>>$",
                parser_state.token_stack,
            )
            POGGER.debug("token_document>>$", parser_state.token_document)
            (
                container_level_tokens,
                requeue_line_info,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[
                    ParagraphStackToken,
                    IndentedCodeBlockStackToken,
                    LinkDefinitionStackToken,
                    HtmlBlockStackToken,
                ],
                was_forced=True,
                caller_can_handle_requeue=True,
            )
            if requeue_line_info and requeue_line_info.lines_to_requeue:
                # TODO is this common?
                POGGER.debug(
                    "__ensure_stack_at_level>>lines_to_requeue>>$",
                    requeue_line_info.lines_to_requeue,
                )
                POGGER.debug(
                    "__close_required_lists_after_start>>parser_state.original_line_to_parse>>$",
                    parser_state.original_line_to_parse,
                )
                POGGER.debug(
                    "__ensure_stack_at_level>>token_stack>>$",
                    parser_state.token_stack,
                )
                POGGER.debug(
                    "__ensure_stack_at_level>>token_document>>$",
                    parser_state.token_document,
                )
                assert not requeue_line_info.lines_to_requeue[0]
                requeue_line_info.lines_to_requeue[
                    0
                ] = parser_state.original_line_to_parse
                POGGER.debug(
                    "__close_required_lists_after_start>>lines_to_requeue>>$",
                    requeue_line_info.lines_to_requeue,
                )
                return None, None, requeue_line_info
            POGGER.debug("token_stack>>$", parser_state.token_stack)
            POGGER.debug("token_document>>$", parser_state.token_document)
            POGGER.debug(
                "container_level_tokens>>$",
                container_level_tokens,
            )
            keep_going = True
            while keep_going and parser_state.token_stack[-1].is_list:
                POGGER.debug("stack>>$", parser_state.token_stack[-1].indent_level)
                POGGER.debug("original_start_index>>$", original_start_index)

                if original_start_index < parser_state.token_stack[-1].indent_level:
                    close_tokens, _ = parser_state.close_open_blocks_fn(
                        parser_state,
                        include_lists=True,
                        was_forced=True,
                        until_this_index=len(parser_state.token_stack) - 1,
                    )
                    container_level_tokens.extend(close_tokens)
                    POGGER.debug("container_level_tokens>>$", container_level_tokens)
                else:
                    keep_going = False

            POGGER.debug("container_level_tokens>>$", container_level_tokens)
            while this_bq_count > stack_bq_count:
                POGGER.debug(
                    "increasing block quotes by one>>",
                )
                stack_bq_count += 1

                adjusted_position_marker = PositionMarker(
                    position_marker.line_number,
                    original_start_index,
                    position_marker.text_to_parse,
                )

                assert (
                    position_marker.text_to_parse[original_start_index]
                    == BlockQuoteProcessor.__block_quote_character
                )
                original_start_index += 1
                if ParserHelper.is_character_at_index_whitespace(
                    position_marker.text_to_parse, original_start_index
                ):
                    original_start_index += 1

                new_markdown_token = BlockQuoteMarkdownToken(
                    extracted_whitespace, adjusted_position_marker
                )

                container_level_tokens.append(new_markdown_token)
                parser_state.token_stack.append(
                    BlockQuoteStackToken(new_markdown_token)
                )
            POGGER.debug("container_level_tokens>>$", container_level_tokens)
            stack_bq_count = BlockQuoteProcessor.__decrease_stack_to_level(
                parser_state, this_bq_count, stack_bq_count, container_level_tokens
            )
            POGGER.debug(
                "container_level_tokens>>$",
                container_level_tokens,
            )

        return container_level_tokens, stack_bq_count, None

    @staticmethod
    def __decrease_stack_to_level(
        parser_state, this_bq_count, stack_bq_count, container_level_tokens
    ):
        while this_bq_count < stack_bq_count:
            POGGER.debug(
                "decreasing block quotes by one>>",
            )
            stack_bq_count -= 1
            (new_tokens, _,) = parser_state.close_open_blocks_fn(
                parser_state,
                include_block_quotes=True,
                until_this_index=len(parser_state.token_stack) - 1,
                was_forced=True,
            )
            container_level_tokens.extend(new_tokens)
        return stack_bq_count

    # pylint: enable=too-many-arguments, too-many-statements, too-many-locals



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_445():
    """
    Test case 445:  (part 1) Rule 11
    """

    # Arrange
    source_markdown = """foo ***"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[text(1,5):***:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo ***</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_446():
    """
    Test case 446:  (part 2) Rule 11
    """

    # Arrange
    source_markdown = """foo *\\**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):1:*]",
        "[text(1,6):\\\b*:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <em>*</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_447():
    """
    Test case 447:  (part 3) Rule 11
    """

    # Arrange
    source_markdown = """foo *_*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):1:*]",
        "[text(1,6):_:]",
        "[end-emphasis(1,7)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <em>_</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_448():
    """
    Test case 448:  (part 4) Rule 11
    """

    # Arrange
    source_markdown = """foo *****"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[text(1,5):*****:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo *****</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_449():
    """
    Test case 449:  (part 5) Rule 11
    """

    # Arrange
    source_markdown = """foo **\\***"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):2:*]",
        "[text(1,7):\\\b*:]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <strong>*</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_450():
    """
    Test case 450:  (part 6) Rule 11
    """

    # Arrange
    source_markdown = """foo **_**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):2:*]",
        "[text(1,7):_:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <strong>_</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_451():
    """
    Test case 451:  (part 1) Note that when delimiters do not match evenly, Rule 11 determines that the excess literal * characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """**foo*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[emphasis(1,2):1:*]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*<em>foo</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_452():
    """
    Test case 452:  (part 2) Note that when delimiters do not match evenly, Rule 11 determines that the excess literal * characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """*foo**"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo:]",
        "[end-emphasis(1,5)::]",
        "[text(1,6):*:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo</em>*</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_453():
    """
    Test case 453:  (part 3) Note that when delimiters do not match evenly, Rule 11 determines that the excess literal * characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """***foo**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[emphasis(1,2):2:*]",
        "[text(1,4):foo:]",
        "[end-emphasis(1,7)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*<strong>foo</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_454():
    """
    Test case 454:  (part 4) Note that when delimiters do not match evenly, Rule 11 determines that the excess literal * characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """****foo*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):***:]",
        "[emphasis(1,4):1:*]",
        "[text(1,5):foo:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>***<em>foo</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_455():
    """
    Test case 455:  (part 5) Note that when delimiters do not match evenly, Rule 11 determines that the excess literal * characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """**foo***"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):2:*]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[text(1,8):*:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><strong>foo</strong>*</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_456():
    """
    Test case 456:  (part 6) Note that when delimiters do not match evenly, Rule 11 determines that the excess literal * characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """*foo****"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo:]",
        "[end-emphasis(1,5)::]",
        "[text(1,6):***:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo</em>***</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module that provides for an encapsulation of properties for an application.
"""
import copy
import json
import logging

LOGGER = logging.getLogger(__name__)


class ApplicationProperties:
    """
    Class that provides for an encapsulation of properties for an application.

    Eventually want to add:
    - set from command line
    - exposure to command line: i.e. list all properties, values, etc.
    - transformers, both per call and registered, to change from one value to another
      - i.e. what if a property is given on the command line as a string, but an integer is required?
    """

    __separator = "."
    __assignment_operator = "="
    __manual_property_type_prefix = "$"
    __manual_property_type_string = "$"
    __manual_property_type_integer = "#"
    __manual_property_type_boolean = "!"

    """
    Class to provide for a container of properties that belong to the application.
    """

    def __init__(self, strict_mode=False):
        """
        Initializes an new instance of the ApplicationProperties class.
        """
        self.__flat_property_map = {}
        self.__strict_mode = strict_mode

    @property
    def separator(self):
        """
        Separator used to split the hierarchy of the property names.
        """
        return self.__separator

    @property
    def number_of_properties(self):
        """
        Number of properties that exist in the map.
        """
        return len(self.__flat_property_map)

    @property
    def property_names(self):
        """
        List of each of the properties in the map.
        """
        return self.__flat_property_map.keys()

    @property
    def strict_mode(self):
        """
        Gets whether strict mode is on by default.
        """
        return self.__strict_mode

    def enable_strict_mode(self):
        """
        Sets struct mode to True to enable it.
        """
        self.__strict_mode = True

    def load_from_dict(self, config_map):
        """
        Load the properties from a provided dictionary.
        """

        if not isinstance(config_map, dict):
            raise ValueError("Specified parameter was not a dictionary.")

        LOGGER.debug("Loading from dictionary: {%s}", str(config_map))
        self.__flat_property_map.clear()
        self.__scan_map(config_map, "")

    @staticmethod
    def verify_full_part_form(property_key):
        """
        Given one part of a full key, verify that it is composed properly.
        """

        if (
            " " in property_key
            or "\t" in property_key
            or "\n" in property_key
            or ApplicationProperties.__assignment_operator in property_key
        ):
            raise ValueError(
                f"Each part of the property key must not contain a whitespace character or the '{ApplicationProperties.__separator}' character."
            )
        if not property_key:
            raise ValueError(
                "Each part of the property key must contain at least one character."
            )
        return property_key

    @staticmethod
    def verify_full_key_form(property_key):
        """
        Given a full key, verify that it is composed properly.
        """

        if property_key.startswith(
            ApplicationProperties.__separator
        ) or property_key.endswith(ApplicationProperties.__separator):
            raise ValueError(
                f"Full property key must not start or end with the '{ApplicationProperties.__separator}' character."
            )
        doubles = (
            f"{ApplicationProperties.__separator}{ApplicationProperties.__separator}"
        )
        doubles_index = property_key.find(doubles)
        if doubles_index != -1:
            raise ValueError(
                f"Full property key cannot contain multiples of the {ApplicationProperties.__separator} without any text between them."
            )
        split_key = property_key.split(ApplicationProperties.__separator)
        for next_key in split_key:
            ApplicationProperties.verify_full_part_form(next_key)
        return property_key

    @staticmethod
    def verify_manual_property_form(string_to_verify):
        """
        Verify the general form of a manual property string. i.e. key=value
        """

        if not isinstance(string_to_verify, str):
            raise ValueError("Manual property form must be a string.")
        equals_index = string_to_verify.find(
            ApplicationProperties.__assignment_operator
        )
        if equals_index == -1:
            raise ValueError(
                "Manual property key and value must be separated by the '=' character."
            )
        property_key = string_to_verify[0:equals_index]
        ApplicationProperties.verify_full_key_form(property_key)
        return string_to_verify

    def set_manual_property(self, combined_string):
        """
        Manually set a property for the object.
        """

        if not isinstance(combined_string, str):
            iterator = None
            try:
                iterator = iter(combined_string)
            except TypeError as this_exception:
                raise ValueError(
                    "Manual property form must either be a string or an iterable of strings."
                ) from this_exception
            for i in iterator:
                self.set_manual_property(i)
            return

        ApplicationProperties.verify_manual_property_form(combined_string)
        equals_index = combined_string.find(ApplicationProperties.__assignment_operator)
        property_key = combined_string[0:equals_index].lower()
        property_value = combined_string[equals_index + 1 :]

        if (
            property_value.startswith(
                ApplicationProperties.__manual_property_type_prefix
            )
            and len(property_value) >= 2
        ):
            if property_value[1] == ApplicationProperties.__manual_property_type_string:
                property_value = property_value[2:]
            elif (
                property_value[1]
                == ApplicationProperties.__manual_property_type_integer
            ):
                try:
                    property_value = int(property_value[2:])
                except ValueError as this_exception:
                    raise ValueError(
                        f"Manual property value '{property_value}' cannot be translated into an integer."
                    ) from this_exception
            elif (
                property_value[1]
                == ApplicationProperties.__manual_property_type_boolean
            ):
                property_value = property_value[2:].lower() == "true"
            else:
                property_value = property_value[1:]
        self.__flat_property_map[property_key] = copy.deepcopy(property_value)
        LOGGER.debug(
            "Adding configuration '%s' : {%s}", property_key, str(property_value)
        )

    # pylint: disable=unidiomatic-typecheck
    # pylint: disable=too-many-arguments
    # pylint: disable=broad-except
    # pylint: disable=raise-missing-from
    def get_property(
        self,
        property_name,
        property_type,
        default_value=None,
        valid_value_fn=None,
        is_required=False,
        strict_mode=None,
    ):
        """
        Get an property of a generic type from the configuration.
        """

        if strict_mode is None:
            strict_mode = self.__strict_mode

        if not isinstance(property_name, str):
            raise ValueError("The propertyName argument must be a string.")
        ApplicationProperties.verify_full_key_form(property_name)
        if not isinstance(property_type, type):
            raise ValueError(
                f"The property_type argument for '{property_name}' must be a type."
            )
        if default_value is not None and type(default_value) != property_type:
            raise ValueError(
                f"The default value for property '{property_name}' must either be None or a '{property_type.__name__}' value."
            )

        property_value = default_value
        property_name = property_name.lower()
        LOGGER.debug("property_name=%s", property_name)
        if property_name in self.__flat_property_map:
            found_value = self.__flat_property_map[property_name]
            is_eligible = type(found_value) == property_type
            if not is_eligible and strict_mode:
                raise ValueError(
                    f"The value for property '{property_name}' must be of type '{property_type.__name__}'."
                )
            if is_eligible and valid_value_fn:
                try:
                    valid_value_fn(found_value)
                except Exception as this_exception:
                    is_eligible = False
                    if strict_mode:
                        raise ValueError(
                            f"The value for property '{property_name}' is not valid: {str(this_exception)}"
                        )
            if is_eligible:
                property_value = found_value
        elif is_required:
            raise ValueError(
                f"A value for property '{property_name}' must be provided."
            )
        return property_value

    # pylint: enable=unidiomatic-typecheck
    # pylint: enable=too-many-arguments
    # pylint: enable=broad-except
    # pylint: enable=raise-missing-from

    def get_boolean_property(
        self, property_name, default_value=None, is_required=False, strict_mode=None
    ):
        """
        Get a boolean property from the configuration.
        """
        return self.get_property(
            property_name,
            bool,
            default_value=default_value,
            valid_value_fn=None,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: disable=too-many-arguments
    def get_integer_property(
        self,
        property_name,
        default_value=None,
        valid_value_fn=None,
        is_required=False,
        strict_mode=None,
    ):
        """
        Get an integer property from the configuration.
        """
        return self.get_property(
            property_name,
            int,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def get_string_property(
        self,
        property_name,
        default_value=None,
        valid_value_fn=None,
        is_required=False,
        strict_mode=None,
    ):
        """
        Get a string property from the configuration.
        """
        return self.get_property(
            property_name,
            str,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    def __scan_map(self, config_map, current_prefix):
        for next_key in config_map:
            if not isinstance(next_key, str):
                raise ValueError(
                    "All keys in the main dictionary and nested dictionaries must be strings."
                )
            if self.__separator in next_key:
                raise ValueError(
                    f"Keys strings cannot contain the separator character '{self.__separator}'."
                )

            next_value = config_map[next_key]
            if isinstance(next_value, dict):
                self.__scan_map(
                    next_value, f"{current_prefix}{next_key}{self.__separator}"
                )
            else:
                new_key = f"{current_prefix}{next_key}".lower()
                self.__flat_property_map[new_key] = copy.deepcopy(next_value)
                LOGGER.debug(
                    "Adding configuration '%s' : {%s}", new_key, str(next_value)
                )


class ApplicationPropertiesFacade:
    """
    Class to provide for a facade in front of an ApplicationProperties instance that
    only exposes part of the properties tree.
    """

    def __init__(self, base_properties, property_prefix):
        """
        Initializes an new instance of the ApplicationPropertiesFacade class.
        """
        if not isinstance(base_properties, ApplicationProperties):
            raise ValueError(
                "The base_properties of the facade must be an ApplicationProperties instance."
            )
        self.__base_properties = base_properties

        if not isinstance(property_prefix, str):
            raise ValueError("The property_prefix argument must be a string.")
        if not property_prefix.endswith(base_properties.separator):
            raise ValueError(
                f"The property_prefix argument must end with the separator character '{base_properties.separator}'."
            )
        self.__property_prefix = property_prefix

    # pylint: disable=too-many-arguments
    def get_property(
        self,
        property_name,
        property_type,
        default_value=None,
        valid_value_fn=None,
        is_required=False,
        strict_mode=False,
    ):
        """
        Get an property of a generic type from the configuration.
        """

        return self.__base_properties.get_property(
            f"{self.__property_prefix}{property_name}",
            property_type,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    def get_boolean_property(
        self, property_name, default_value=None, is_required=False
    ):
        """
        Get a boolean property from the configuration.
        """
        return self.__base_properties.get_boolean_property(
            f"{self.__property_prefix}{property_name}",
            default_value=default_value,
            is_required=is_required,
        )

    # pylint: disable=too-many-arguments
    def get_integer_property(
        self,
        property_name,
        default_value=None,
        valid_value_fn=None,
        is_required=False,
        strict_mode=None,
    ):
        """
        Get an integer property from the configuration.
        """
        return self.__base_properties.get_integer_property(
            f"{self.__property_prefix}{property_name}",
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def get_string_property(
        self,
        property_name,
        default_value=None,
        valid_value_fn=None,
        is_required=False,
        strict_mode=None,
    ):
        """
        Get a string property from the configuration.
        """
        return self.__base_properties.get_string_property(
            f"{self.__property_prefix}{property_name}",
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    @property
    def property_names(self):
        """
        List of each of the properties in the map.
        """
        facade_property_names = []
        for next_property_name in self.__base_properties.property_names:
            if next_property_name.startswith(self.__property_prefix):
                facade_property_names.append(
                    next_property_name[len(self.__property_prefix) :]
                )
        return facade_property_names


# pylint: disable=too-few-public-methods
class ApplicationPropertiesJsonLoader:
    """
    Class to provide for a manner to load an ApplicationProperties object from a JSON file.
    """

    @staticmethod
    def load_and_set(properties_object, configuration_file, handle_error_fn):
        """
        Load the specified file and set it into the given properties object.
        """

        try:
            with open(configuration_file) as infile:
                configuration_map = json.load(infile)
        except json.decoder.JSONDecodeError as this_exception:
            formatted_error = f"Specified configuration file '{configuration_file}' is not a valid JSON file ({str(this_exception)})."
            handle_error_fn(formatted_error, this_exception)
        except IOError as this_exception:
            formatted_error = f"Specified configuration file '{configuration_file}' was not loaded ({str(this_exception)})."
            handle_error_fn(formatted_error, this_exception)

        try:
            properties_object.load_from_dict(configuration_map)
        except ValueError as this_exception:
            formatted_error = f"Specified configuration file '{configuration_file}' is not valid ({str(this_exception)})."
            handle_error_fn(formatted_error, this_exception)


# pylint: enable=too-few-public-methods



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_387():
    """
    Test case 387:  Rule 5:
    """

    # Arrange
    source_markdown = """**foo bar**"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):2:*]",
        "[text(1,3):foo bar:]",
        "[end-emphasis(1,10)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><strong>foo bar</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_388():
    """
    Test case 388:  This is not strong emphasis, because the opening delimiter is followed by whitespace:
    """

    # Arrange
    source_markdown = """** foo bar**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):**:]",
        "[text(1,3): foo bar:]",
        "[text(1,11):**:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>** foo bar**</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_389():
    """
    Test case 389:  This is not strong emphasis, because the opening ** is preceded by an alphanumeric and followed by punctuation, and hence not part of a left-flanking delimiter run:
    """

    # Arrange
    source_markdown = """a**"foo"**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a:]",
        "[text(1,2):**:]",
        '[text(1,4):\a"\a&quot;\afoo\a"\a&quot;\a:]',
        "[text(1,9):**:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a**&quot;foo&quot;**</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_390():
    """
    Test case 390:  Intraword strong emphasis with ** is permitted:
    """

    # Arrange
    source_markdown = """foo**bar**"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo:]",
        "[emphasis(1,4):2:*]",
        "[text(1,6):bar:]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo<strong>bar</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_paragraph_series_h_f_l_t():
    """
    Test case:  Paragraph with full link label text split over 2 lines
    was:        test_paragraph_extra_a6
    """

    # Arrange
    source_markdown = """a[li
nk][bar]a

[bar]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):full:/url:title:::bar:li\nnk:::::]",
        "[text(1,3):li\nnk::\n]",
        "[end-link::]",
        "[text(2,9):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar:: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">li\nnk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_f_i_t():
    """
    Test case:  Paragraph with full image label text split over 2 lines
    was:        test_paragraph_extra_b8
    """

    # Arrange
    source_markdown = """a![li
nk][bar]a

[bar]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):full:/url:title:li\nnk:::bar:li\nnk:::::]",
        "[text(2,9):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar:: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li\nnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_f_l_cs():
    """
    Test case:  Paragraph with full link label code span split over 2 lines
    was:        test_paragraph_extra_a7
    """

    # Arrange
    source_markdown = """a[li`de
fg`nk][bar]a

[bar]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):full:/url:title:::bar:li`de\nfg`nk:::::]",
        "[text(1,3):li:]",
        "[icode-span(1,5):de\a\n\a \afg:`::]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,12):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar:: :/url:: :title:'title':]",
    ]
    expected_gfm = (
        """<p>a<a href="/url" title="title">li<code>de fg</code>nk</a>a</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_f_i_cs():
    """
    Test case:  Paragraph with full image label code span split over 2 lines
    was:        test_paragraph_extra_b9
    """

    # Arrange
    source_markdown = """a![li`de
fg`nk][bar]a

[bar]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):full:/url:title:lide fgnk:::bar:li`de\nfg`nk:::::]",
        "[text(2,12):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar:: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="lide fgnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_f_l_rh():
    """
    Test case:  Paragraph with full link label raw html split over 2 lines
    was:        test_paragraph_extra_a8
    """

    # Arrange
    source_markdown = """a[li<de
fg>nk][bar]a

[bar]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):full:/url:title:::bar:li<de\nfg>nk:::::]",
        "[text(1,3):li:]",
        "[raw-html(1,5):de\nfg]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,12):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar:: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">li<de\nfg>nk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_f_i_rh():
    """
    Test case:  Paragraph with full image label raw html split over 2 lines
    was:        test_paragraph_extra_c0
    """

    # Arrange
    source_markdown = """a![li<de
fg>nk][bar]a

[bar]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):full:/url:title:li<de\nfg>nk:::bar:li<de\nfg>nk:::::]",
        "[text(2,12):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::bar:: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li<de\nfg>nk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_c_l_t():
    """
    Test case:  Paragraph with collapsed link label text split over 2 lines
    was:        test_paragraph_extra_a9
    """

    # Arrange
    source_markdown = """a[li
nk][]a

[li\nnk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):collapsed:/url:title::::li\nnk:::::]",
        "[text(1,3):li\nnk::\n]",
        "[end-link::]",
        "[text(2,6):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li nk:li\nnk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">li\nnk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_c_i_t():
    """
    Test case:  Paragraph with collapsed image label text split over 2 lines
    was:        test_paragraph_extra_c1
    """

    # Arrange
    source_markdown = """a![li
nk][]a

[li\nnk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):collapsed:/url:title:li\nnk::::li\nnk:::::]",
        "[text(2,6):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li nk:li\nnk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li\nnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_c_l_cs():
    """
    Test case:  Paragraph with collapsed link label code span split over 2 lines
    was:        test_paragraph_extra_b0
    """

    # Arrange
    source_markdown = """a[li`de
fg`nk][]a

[li`de\nfg`nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):collapsed:/url:title::::li`de\nfg`nk:::::]",
        "[text(1,3):li:]",
        "[icode-span(1,5):de\a\n\a \afg:`::]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,9):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li`de fg`nk:li`de\nfg`nk: :/url:: :title:'title':]",
    ]
    expected_gfm = (
        """<p>a<a href="/url" title="title">li<code>de fg</code>nk</a>a</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_c_i_cs():
    """
    Test case:  Paragraph with collapsed image label code span split over 2 lines
    was:        test_paragraph_extra_c2
    """

    # Arrange
    source_markdown = """a![li`de
fg`nk][]a

[li`de\nfg`nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):collapsed:/url:title:lide fgnk::::li`de\nfg`nk:::::]",
        "[text(2,9):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li`de fg`nk:li`de\nfg`nk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="lide fgnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_c_l_rh():
    """
    Test case:  Paragraph with collapsed link label raw html split over 2 lines
    was:        test_paragraph_extra_b1
    """

    # Arrange
    source_markdown = """a[li<de
fg>nk][]a

[li<de\nfg>nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):collapsed:/url:title::::li<de\nfg>nk:::::]",
        "[text(1,3):li:]",
        "[raw-html(1,5):de\nfg]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,9):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li<de fg>nk:li<de\nfg>nk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">li<de\nfg>nk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_c_i_rh():
    """
    Test case:  Paragraph with collapsed image label raw html split over 2 lines
    was:        test_paragraph_extra_c3
    """

    # Arrange
    source_markdown = """a![li<de
fg>nk][]a

[li<de\nfg>nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):collapsed:/url:title:li<de\nfg>nk::::li<de\nfg>nk:::::]",
        "[text(2,9):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li<de fg>nk:li<de\nfg>nk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li<de\nfg>nk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_s_l_t():
    """
    Test case:  Paragraph with shortcut link label text split over 2 lines
    was:        test_paragraph_extra_b2
    """

    # Arrange
    source_markdown = """a[li
nk]a

[li\nnk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):shortcut:/url:title::::li\nnk:::::]",
        "[text(1,3):li\nnk::\n]",
        "[end-link::]",
        "[text(2,4):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li nk:li\nnk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">li\nnk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_s_i_t():
    """
    Test case:  Paragraph with shortcut image label text split over 2 lines
    was:        test_paragraph_extra_c4
    """

    # Arrange
    source_markdown = """a![li
nk]a

[li\nnk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):shortcut:/url:title:li\nnk::::li\nnk:::::]",
        "[text(2,4):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li nk:li\nnk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li\nnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_s_l_cs():
    """
    Test case:  Paragraph with shortcut link label code span split over 2 lines
    was:        test_paragraph_extra_b3
    """

    # Arrange
    source_markdown = """a[li`de
fg`nk]a

[li`de\nfg`nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):shortcut:/url:title::::li`de\nfg`nk:::::]",
        "[text(1,3):li:]",
        "[icode-span(1,5):de\a\n\a \afg:`::]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,7):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li`de fg`nk:li`de\nfg`nk: :/url:: :title:'title':]",
    ]
    expected_gfm = (
        """<p>a<a href="/url" title="title">li<code>de fg</code>nk</a>a</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_s_i_cs():
    """
    Test case:  Paragraph with shortcut image label code span split over 2 lines
    was:        test_paragraph_extra_c5
    """

    # Arrange
    source_markdown = """a![li`de
fg`nk]a

[li`de\nfg`nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):shortcut:/url:title:lide fgnk::::li`de\nfg`nk:::::]",
        "[text(2,7):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li`de fg`nk:li`de\nfg`nk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="lide fgnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_s_l_rh():
    """
    Test case:  Paragraph with shortcut link label raw html split over 2 lines
    was:        test_paragraph_extra_b4
    """

    # Arrange
    source_markdown = """a[li<de
fg>nk]a

[li<de\nfg>nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[link(1,2):shortcut:/url:title::::li<de\nfg>nk:::::]",
        "[text(1,3):li:]",
        "[raw-html(1,5):de\nfg]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,7):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li<de fg>nk:li<de\nfg>nk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<a href="/url" title="title">li<de\nfg>nk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_s_i_rh():
    """
    Test case:  Paragraph with shortcut image label raw html split over 2 lines
    was:        test_paragraph_extra_c6
    """

    # Arrange
    source_markdown = """a![li<de
fg>nk]a

[li<de\nfg>nk]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        "[image(1,2):shortcut:/url:title:li<de\nfg>nk::::li<de\nfg>nk:::::]",
        "[text(2,7):a:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,1):True::li<de fg>nk:li<de\nfg>nk: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p>a<img src="/url" alt="li<de\nfg>nk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_i_l_t():
    """
    Test case:  Paragraph with inline link label text split over 2 lines
    was:        test_paragraph_extra_a3
    """

    # Arrange
    source_markdown = """a[li
nk](/uri "title" )a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::li\nnk:False:":: : ]',
        "[text(1,3):li\nnk::\n]",
        "[end-link::]",
        "[text(2,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">li\nnk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_i_i_t():
    """
    Test case:  Paragraph with inline image label text split over 2 lines
    was:        test_paragraph_extra_b5
    """

    # Arrange
    source_markdown = """a![li\nnk](/uri "title" )a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:li\nnk::::li\nnk:False:":: : ]',
        "[text(2,19):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="li\nnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_i_l_cs():
    """
    Test case:  Paragraph with inline link label code span split over 2 lines
    was:        test_paragraph_extra_a4
    """

    # Arrange
    source_markdown = """a[li`de
fg`nk](/uri "title" )a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::li`de\nfg`nk:False:":: : ]',
        "[text(1,3):li:]",
        "[icode-span(1,5):de\a\n\a \afg:`::]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,22):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>a<a href="/uri" title="title">li<code>de fg</code>nk</a>a</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_i_i_cs():
    """
    Test case:  Paragraph with inline image label code span split over 2 lines
    was:        test_paragraph_extra_b6
    """

    # Arrange
    source_markdown = """a![li`de\nfg`nk](/uri "title" )a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:lide fgnk::::li`de\nfg`nk:False:":: : ]',
        "[text(2,22):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="lide fgnk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_i_l_rh():
    """
    Test case:  Paragraph with inline link label raw html split over 2 lines
    was:        test_paragraph_extra_a5
    """

    # Arrange
    source_markdown = """a[li<de
fg>nk](/uri "title" )a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[link(1,2):inline:/uri:title::::li<de\nfg>nk:False:":: : ]',
        "[text(1,3):li:]",
        "[raw-html(1,5):de\nfg]",
        "[text(2,4):nk:]",
        "[end-link::]",
        "[text(2,22):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri" title="title">li<de\nfg>nk</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_h_i_i_rh():
    """
    Test case:  Paragraph with inline image label raw html split over 2 lines
    was:        test_paragraph_extra_b7
    """

    # Arrange
    source_markdown = """a![li<de
fg>nk](/uri "title" )a"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):a:]",
        '[image(1,2):inline:/uri:title:li<de\nfg>nk::::li<de\nfg>nk:False:":: : ]',
        "[text(2,22):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<img src="/uri" alt="li<de\nfg>nk" title="title" />a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#emphasis-and-strong-emphasis
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_emphasis_457():
    """
    Test case 457:  (part 1) Rule 12
    """

    # Arrange
    source_markdown = """foo ___"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[text(1,5):___:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo ___</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_458():
    """
    Test case 458:  (part 2) Rule 12
    """

    # Arrange
    source_markdown = """foo _\\__"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):1:_]",
        "[text(1,6):\\\b_:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <em>_</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_459():
    """
    Test case 459:  (part 3) Rule 12
    """

    # Arrange
    source_markdown = """foo _*_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):1:_]",
        "[text(1,6):*:]",
        "[end-emphasis(1,7)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <em>*</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_460():
    """
    Test case 460:  (part 4) Rule 12
    """

    # Arrange
    source_markdown = """foo _____"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[text(1,5):_____:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo _____</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_461():
    """
    Test case 461:  (part 5) Rule 12
    """

    # Arrange
    source_markdown = """foo __\\___"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):2:_]",
        "[text(1,7):\\\b_:]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <strong>_</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_462():
    """
    Test case 462:  (part 6) Rule 12
    """

    # Arrange
    source_markdown = """foo __*__"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):foo :]",
        "[emphasis(1,5):2:_]",
        "[text(1,7):*:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>foo <strong>*</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_463():
    """
    Test case 463:  (part 7) Rule 12
    """

    # Arrange
    source_markdown = """__foo_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[emphasis(1,2):1:_]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_<em>foo</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_464():
    """
    Test case 464:  (part 1) Note that when delimiters do not match evenly, Rule 12 determines that the excess literal _ characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """_foo__"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):foo:]",
        "[end-emphasis(1,5)::]",
        "[text(1,6):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo</em>_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_465():
    """
    Test case 465:  (part 2) Note that when delimiters do not match evenly, Rule 12 determines that the excess literal _ characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """___foo__"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[emphasis(1,2):2:_]",
        "[text(1,4):foo:]",
        "[end-emphasis(1,7)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_<strong>foo</strong></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_466():
    """
    Test case 466:  (part 3) Note that when delimiters do not match evenly, Rule 12 determines that the excess literal _ characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """____foo_"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):___:]",
        "[emphasis(1,4):1:_]",
        "[text(1,5):foo:]",
        "[end-emphasis(1,8)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>___<em>foo</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_467():
    """
    Test case 467:  (part 4) Note that when delimiters do not match evenly, Rule 12 determines that the excess literal _ characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """__foo___"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):2:_]",
        "[text(1,3):foo:]",
        "[end-emphasis(1,6)::]",
        "[text(1,8):_:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><strong>foo</strong>_</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_emphasis_468():
    """
    Test case 468:  (part 5) Note that when delimiters do not match evenly, Rule 12 determines that the excess literal _ characters will appear outside of the emphasis, rather than inside it:
    """

    # Arrange
    source_markdown = """_foo____"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:_]",
        "[text(1,2):foo:]",
        "[end-emphasis(1,5)::]",
        "[text(1,6):___:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo</em>___</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
# pylint: disable=wrong-import-position
# pylint: disable=consider-using-enumerate
# pylint: disable=no-self-use
# pylint: disable=wrong-import-order
import timeit

import pytest

from pymarkdown.parser_helper import ParserHelper


def xtest_block_quotes_extra_perf1():

    n = 500
    ex = (
        "\n"
        + "from pymarkdown.tokenized_markdown import TokenizedMarkdown\n"
        + "source_markdown = '> a``html\\n> maybe``\\n>\\n> end'\n"
        + "tokenizer = TokenizedMarkdown()\n"
        + "actual_tokens = tokenizer.transform(source_markdown, show_debug=False)\n"
    )
    n = 1
    ex = (
        "import logging\n"
        + "LOGGER = logging.getLogger(__name__)\n"
        + "logging.getLogger().setLevel(logging.WARNING)\n"
        + "for i in range(0,100000):\n"
        + "   LOGGER.debug('this is something to log1-->')\n"
    )

    x = timeit.timeit(ex, number=n)
    print(">>" + str(x))
    print(">>" + str(float(x) / float(n)))
    assert False


def bob1():
    search_string = "1234567890z"
    search_for = "abcdefghijklmnopqrstuvwxyz"
    ParserHelper.index_any_of(search_string, search_for)


def fred():
    f = "this is a test!"
    g = "1"
    h = "yyyyy"
    hh = "yyyyy33"
    i = "".join([f, g, h, hh])
    return i


def barney():
    f = "this is a test!"
    g = "1"
    h = "yyyyy"
    hh = "yyyyy33"
    i = f"{f}{g}{h}{hh}"
    return i


def xtest_block_quotes_extra_perf3x():
    a = "a"
    b = 1
    c = f"{a}{b}"
    assert c == "a1"


@pytest.mark.timeout(120)
def xtest_block_quotes_extra_perf3():

    assert fred() == barney()
    for y in range(0, 5):
        xx = timeit.timeit("barney()", globals=globals(), number=100)
    for y in range(0, 5):
        xx = timeit.timeit("fred()", globals=globals(), number=100)

    for y in range(0, 5):
        xx = timeit.timeit("barney()", globals=globals(), number=10000000)
        print(str(xx) + ">>" + str(y))
    for y in range(0, 5):
        xx = timeit.timeit("fred()", globals=globals(), number=10000000)
        print(str(xx) + ">>" + str(y))
    assert False


# pylint: enable=missing-module-docstring
# pylint: enable=missing-class-docstring
# pylint: enable=missing-function-docstring
# pylint: enable=invalid-name
# pylint: enable=wrong-import-position
# pylint: enable=consider-using-enumerate
# pylint: enable=no-self-use
# pylint: enable=wrong-import-order
